import json
import os
import time

from google.cloud.monitoring_v3 import MetricServiceClient, TimeInterval, ListTimeSeriesRequest  # noqa: E501
from google.cloud import storage


def getBlocks(sub_id: str, path_to_output: str = "./", dry=False):
    if not dry:
        # Instantiates a client
        storage_client = storage.Client(project=sub_id)

        # List all the buckets available
        client = MetricServiceClient()
        # for bucket in storage_client.list_buckets():
        #     print(bucket)
        #     print(bucket.name)

        now = time.time()
        seconds = int(now)
        nanos = int((now - seconds) * 10**9)
        interval = TimeInterval(
            {
                "end_time": {"seconds": seconds, "nanos": nanos},
                "start_time": {"seconds": (seconds - 300), "nanos": nanos},
            }
        )

        results = client.list_time_series(
            request={
                "name": f"projects/{sub_id}",
                "filter": 'metric.type = "storage.googleapis.com/storage/total_bytes"',  # noqa: E501
                "interval": interval,
                "view": ListTimeSeriesRequest.TimeSeriesView.FULL,
            }
        )
        my_buckets = []
        known_buckets = {}
        all_buckets = storage_client.list_buckets()
        for bucket in all_buckets:
            rp = bucket.retention_period
            known_buckets[bucket.name] = {
                "name": bucket.name,
                "size": 0,
                "retention": rp,
                "public": False
            }
            iam = bucket.get_iam_policy()
            permissions = []
            for b in iam.bindings:
                p = {"permission": b["role"], "roles": list(b["members"])}
                permissions.append(json.dumps(p))
                if "allUsers" in b["members"]:
                    known_buckets[bucket.name]["public"] = True
            known_buckets[bucket.name]["permissions"] = permissions

        for result in results:
            if result.resource and result.resource.labels:
                bn = result.resource.labels["bucket_name"]
                size = result.points[-1].value.double_value
                if bn in known_buckets:
                    known_buckets[bn]["size"] = size
        my_buckets = list(known_buckets.values())
        upload = {
                    "metadata": {
                        "provider": "gcp",
                        "account": str(sub_id)
                    },
                    "data": my_buckets
                }
    # End dry block
    gcp_output_file = os.path.join(
        path_to_output,
        f'gcp-storage-{sub_id}.json'
    )
    if not dry:
        with open(gcp_output_file, 'w') as outfile:
            outfile.write(json.dumps(upload, indent=4))
    return gcp_output_file

# Test Code
# getBlocks(sub_id="startupos-328814")



import json
from pathlib import Path

from verinfast.config import Config
"""
    dry
    should_upload
    base_url
    uuid
    path
    noGit
"""

file_path = Path(__file__)
test_folder = file_path.parent.absolute()


def test_arg_dry():
    c = Config()
    parser = c.init_argparse()
    args = parser.parse_args(['--dry'])
    assert args.dry is True
    c.handle_args(args)
    assert c.dry is True


def test_arg_should_upload():
    c = Config()
    parser = c.init_argparse()
    args = parser.parse_args(['--should_upload'])
    assert args.should_upload is True
    c.handle_args(args)
    assert c.shouldUpload is True


def test_is_remote():
    c = Config()
    c.cfg_path = 'http://www.google.com'
    assert c.is_path_remote() is True
    c.cfg_path = 'https://www.google.com'
    assert c.is_path_remote() is True
    c.cfg_path = 'config.yaml'
    assert c.is_path_remote() is False
    c.cfg_path = './config.yaml'
    assert c.is_path_remote() is False
    c.cfg_path = '//config.yaml'
    assert c.is_path_remote() is False
    c.cfg_path = '/C://config.yaml'
    assert c.is_path_remote() is False


def test_config_to_json():
    c = Config()
    s = str(c)
    d = json.loads(s)
    assert d["cfg_path"] is not None


def test_should_git_with_args():
    conf_path = test_folder.joinpath('str_conf.yaml').absolute()
    config = Config(str(conf_path))
    parser = config.init_argparse()
    args = parser.parse_args(['--truncate_findings=30'])
    assert args.truncate_findings == 30
    config.handle_args(args)
    assert config.runGit is True


def test_should_git_with_args_set():
    config = Config()
    parser = config.init_argparse()
    args = parser.parse_args(['-g'])
    assert args.should_git is True
    config.handle_args(args)
    assert config.runGit is True



# stdlib
import argparse
from dataclasses import dataclass, field, is_dataclass, asdict
from datetime import date
import json
from typing import List, Optional
import os
from pathlib import Path
import sys
from uuid import uuid4

# external
import httpx
import yaml

# internal
from verinfast.utils.utils import DebugLog

default_month_delta = 6

default_end_date = date.today()
default_start_year = default_end_date.year
default_start_month = default_end_date.month-default_month_delta
while default_start_month <= 0:
    default_start_year -= 1
    default_start_month += 12

default_start_date = date(
    year=default_start_year,
    month=default_start_month,
    day=1  # TODO: Support arbitrary start days
)

default_start: str = (
    str(default_start_date.year) + "-" +
    str(default_start_date.month) + "-" +
    str(default_start_date.day)
)
default_end: str = (
    str(default_end_date.year) + "-" +
    str(default_end_date.month) + "-" +
    str(default_end_date.day)
)


class printable:
    def __str__(self):
        d = {}
        for key in dir(self):
            x = self.__getattribute__(key)
            if not key.startswith("_") and not callable(x):  # noqa: E501
                if is_dataclass(x):
                    d[key] = asdict(x)
                elif isinstance(x, date):
                    d[key] = x.strftime('%Y-%mm-%dd')
                elif x is None:
                    d[key] = None
                else:
                    d[key] = x.__str__()

        return json.dumps(d, indent=4, default=str)

    # Dictionary-like access / updates
    # def __getitem__(self, name):
    #     value = self.__dict[name]
    #     if isinstance(value, dict):  # recursively view sub-dicts as objects
    #         value = printable(value)
    #     return value

    # def __setitem__(self, name, value):
    #     self.__dict[name] = value

    # def __delitem__(self, name):
    #     del self.__dict[name]

    # # Object-like access / updates
    # def __getattr__(self, name):
    #     return self[name]

    # def __setattr__(self, name, value):
    #     self[name] = value

    # def __delattr__(self, name):
    #     del self[name]


@dataclass
class UploadConfig(printable):
    """
        Args:
            uuid (bool) : specifies whether to use the uuid path prefix
            prefix (str) : defaults to "/report" if not specified
    """
    uuid: bool = False
    prefix: str | None = "/report/"
    code_separator: str | None = "/CorsisCode"
    cost_separator: str | None = None


@dataclass
class GitModule(printable):
    start: str = default_start


@dataclass
class CodeModule(printable):
    git: GitModule
    dry: bool = False
    dependencies: bool = True


@dataclass
class CloudProvider(printable):
    """Cloud Provider

    This dataclass describes the configuration
    for a cloud provider. Currently AWS, GCP, and Azure
    are supported to various degrees

    Args:
        provider (str): A string representing a provider.
                        Today either 'aws','gcp', or 'azure'.
        account (str | int): The account identifier for the provider.
        profile (str): Profile credentials to use.
        start (str): A date in the format 'YYYY-MM-DD'
        end (str): A date in the format 'YYYY-MM-DD'

    """
    provider: str
    account: str | int
    profile: Optional[str] = None
    start: str = default_start
    end: str = default_end


@dataclass
class ConfigModules(printable):
    code: CodeModule = None
    cloud: List[CloudProvider] = field(default_factory=list)


class Config(printable):
    """VerinFast Config
    VerinFast takes configuration from either a .verinfast.yaml file
    or from the command line via arguments. The program will look for
    a .verinfast.yaml file in the current directory or any ancestor,
    stopping when it finds the first.

    Some values are set in the class definition, and are not generally
    overwriteable. That may change in the future, but for now these
    are intentionally not exposed and are consider for developer use
    only. Examples include: "runGit", "runSizes", etc. Modifying these
    could have unanticipated side effects.
    """
    baseUrl: str = ''
    cfg_path: str = ".verinfast.yaml"
    config = FileNotFoundError
    scanId: int = 0
    delete_config_after = False
    delete_temp = True
    # Flag to not run scans, just upload files (if shouldUpload==True)
    dry: bool = False
    local_scan_path: str = "./"
    modules: ConfigModules | None = None
    output_dir = os.path.join(os.getcwd(), "results")
    reportId: int = 0
    runDependencies: bool = True
    runGit: bool = True
    runScan: bool = True
    runSizes: bool = True
    runStats: bool = True
    server_prefix: str | None = None
    server_code_separator: str | None = None
    server_cost_separator: str | None = None
    shouldUpload: bool = False
    shouldManualFileScan: bool = True
    truncate_findings = False
    truncate_findings_length = 30
    upload_logs = False
    use_uuid = False

    def __init__(self, cfg_path: str = None) -> None:
        if cfg_path is not None:
            self.cfg_path = cfg_path
        elif 'pytest' not in sys.argv[0]:
            parser = self.init_argparse()
            args = parser.parse_args()
            if "config" in args and args.config is not None:
                self.cfg_path = args.config

        # TODO: Support JSON
        if self.is_path_remote():
            self.delete_config_after = True
            requestx = httpx.Client(http2=True, timeout=None)
            response = requestx.get(self.cfg_path)
            self.cfg_path = str(uuid4())+".yaml"
            with open(self.cfg_path, 'wb') as f:
                f.write(response.content)

        else:
            # If passed an argument like config.yaml does it exist here?
            # If not check the parent folder and update our internal path
            if not os.path.isfile(self.cfg_path):
                curr_path = Path(os.getcwd())
                parent = curr_path.parent
                while parent and Path(curr_path) != Path('/'):
                    if curr_path.joinpath(self.cfg_path).exists():
                        self.cfg_path = curr_path.joinpath(self.cfg_path)
                        break
                    parent = curr_path.parent
                    curr_path = parent
        self.handle_config_file()
        if 'pytest' not in sys.argv[0]:
            self.handle_args(args)
        if self.config is FileNotFoundError:
            self.config = {}
        """
        If run with no arguments assume we want to scan the current directory.
        If a scan target is not set by the command line or config we set the
        target to be ["./"]
        """
        if (
            "repos" not in self.config and
            "local_repos" not in self.config and
            (
                "modules" not in self.config or
                "cloud" not in self.config["modules"]
            )
        ):
            self.config["local_repos"] = [self.local_scan_path]
            gm = GitModule()
            cm = CodeModule(git=gm)
            """
                The line below previously overwrote existing modules.
            """
            self.modules = ConfigModules(code=cm, cloud=[])
            self.runGit = False
        elif (
            "repos" in self.config and
            "modules" in self.config and
            "code" in self.config["modules"] and
            "git" in self.config["modules"]["code"]
        ):
            self.runGit = True
        self.upload_conf = UploadConfig(
            uuid=self.use_uuid
        )
        if self.server_cost_separator is not None:
            self.upload_conf.cost_separator = self.server_cost_separator
        if self.server_prefix is not None:
            self.upload_conf.prefix = self.server_prefix
        if self.server_code_separator is not None:
            self.upload_conf.code_separator = self.server_code_separator

        os.makedirs(self.output_dir, exist_ok=True)
        debugLog = DebugLog(path=self.output_dir)
        debugLog.log(msg=self.config, tag="Config", display=True)

    def init_argparse(self) -> argparse.ArgumentParser:
        """config.init_argparse

        Note: Never use default values here. If a value is in the
        arg list it will overwrite the existing value. When a
        default value is used, it always shows up in the arg list
        which means it will overwrite a value already specified in
        a config file, which is not the desired behavior.
        """
        parser = argparse.ArgumentParser(
            prog="verinfast",
            usage="%(prog)s [OPTION] [FILE]...",
            # TODO: Description
            # description="Print or check SHA1 (160-bit) checksums."
        )

        parser.add_argument(
            "-c", "--config",
            dest="config",
            help="""This argument can be used with either a local or
            remote path. A path of 'http://a.b.c/config.yaml' the
            application will download that file, save it, locally,
            parse it, and then delete it when complete.
            It will be saved with a random file name (uuid4) while
            it is on the local machine"""
        )

        parser.add_argument(
            "-o", "--output", "--output_dir",
            dest="output_dir"
        )

        parser.add_argument(
            "-t", "--truncate", "--truncate_findings",
            dest="truncate_findings",
            type=int,
            help="""This flag will further enhance privacy by capping
            The length of security warnings. It defaults to unlimited,
            but can be set to any level you feel comfortable with.

            <0 = unlimited
            We recommend 30 as good balance between privacy and utility
            """
        )

        parser.add_argument(
            "-d", "--dry",
            dest="dry",
            action="store_true",
            help="""This flag is boolean, and when passed will not run
            any scans. It will attempt to upload the results. It is
            primarily used for developing servers that accept these results.

            If this is passed it will set should_upload to true no matter what.
            """
        )

        parser.add_argument(
            "--should_upload",
            action="store_true",
            dest="should_upload",
            help=""" --should_upload tells the application that it should
            send the results to the bsae_url specified either as a command
            line argument or in the config.
            """
        )

        parser.add_argument(
            "--base_url",
            type=str,
            dest="base_url",
            help=""" --base_url=https://a.b.c/ will post the results of the
            scan to that url, provided the server conforms to the expected
            structure and responses."""
        )

        parser.add_argument(
            "--uuid",
            type=str,
            dest="uuid",
            help=""" --uuid specifies the secret key the receiving server
            needs to identify your upload. This can either be read from
            the config file (most common with remote configs), or passed
            here on the command line."""
        )

        parser.add_argument(
            "--path",
            type=str,
            dest="local_scan_path",
            help="""This argument will be ignored if a config file specifies
            repositories to scan. This is only used for a single path scan.
            """
        )

        parser.add_argument(
            "--should_git", "-g",
            action="store_true",
            dest="should_git",
            help="""Used to skip contributions and only run a
            code quality scan."""
        )

        return parser

    def handle_args(self, args):
        """ config.handle_args

        config.handle_args overwrites any values stored in the
        current config with the values passed on the command line.
        """
        print(args)
        if "output_dir" in args and args.output_dir is not None:
            self.output_dir = os.path.join(os.getcwd(), args.output_dir)

        if "uuid" in args and args.uuid is not None:
            self.reportId = args.uuid
            self.use_uuid = True

        if "base_url" in args and args.base_url is not None:
            self.baseUrl = args.base_url

        if "should_upload" in args and args.should_upload:
            self.shouldUpload = True

        if "dry" in args and args.dry:
            self.dry = True

        if "local_scan_path" in args and args.local_scan_path is not None:
            self.local_scan_path = args.local_scan_path

        if "should_git" in args and args.should_git is True:
            self.runGit = args.should_git

        if "truncate_findings" in args and args.truncate_findings is not None:
            if args.truncate_findings >= 0:
                self.truncate_findings = True
                self.truncate_findings_length = args.truncate_findings
            else:
                self.truncate_findings = False

    def is_path_remote(self) -> bool:
        s = self.cfg_path

        # TODO: Add ftp
        supported_protocols = [
            "http",
            "https"
        ]

        # protocol_separator
        ps = "://"
        for sp in supported_protocols:
            if s.lower().startswith(sp+ps):
                return True
        return False

    # TODO: Support JSON
    def handle_config_file(self):
        if os.path.isfile(self.cfg_path):
            # Read the config file
            with open(self.cfg_path) as f:
                self.config = yaml.safe_load(f)

            # Global Configuration
            if "baseurl" in self.config:
                self.baseUrl = self.config["baseurl"]
            if "should_upload" in self.config:
                self.shouldUpload = self.config["should_upload"]
            if "dry" in self.config:
                self.dry = self.config["dry"]
            if "delete_temp" in self.config:
                self.delete_temp = self.config["delete_temp"]
            if "truncate_findings" in self.config:
                self.truncate_findings = self.config["truncate_findings"]
                if "truncate_findings_length" in self.config:
                    self.truncate_findings_length = \
                        self.config["truncate_findings_length"]
                else:
                    self.truncate_findings_length = 30

            if "server" in self.config:
                s = self.config["server"]
                if "prefix" in s:
                    self.server_prefix = s["prefix"]
                if "code_separator" in s:
                    self.server_code_separator = s["code_separator"]
                if "cost_separator" in s:
                    self.server_cost_separator = s["cost_separator"]

            if "report" in self.config:
                if "uuid" in self.config["report"]:
                    self.use_uuid = True
                    self.reportId = self.config["report"]["uuid"]
                elif "id" in self.config["report"]:
                    self.reportId = self.config["report"]["id"]

            # Module specific configuration
            if "modules" in self.config:
                m = self.config["modules"]
                gm = GitModule()
                code_modules = CodeModule(git=gm)
                if "code" in m:
                    c = m["code"]
                    if "run_git" in c:
                        self.runGit = c["run_git"]
                    if "run_scan" in c:
                        self.runScan = c["run_scan"]
                    if "run_sizes" in c:
                        self.runSizes = c["run_sizes"]
                    if "run_stats" in c:
                        self.runStats = c["run_stats"]
                    if "git" in c:
                        g = ["git"]
                        if "start" in g:
                            gm.start = g["start"]
                    if "dependencies" in c:
                        self.runDependencies = c["dependencies"]
                cloud_modules = []
                if "cloud" in m:
                    c = m["cloud"]
                    for row in c:
                        provider = CloudProvider(**row)
                        cloud_modules.append(provider)
                self.modules = ConfigModules(
                    code=code_modules,
                    cloud=cloud_modules
                )



from datetime import datetime
import json
import os
import time
from typing import List

from google.api_core.exceptions import NotFound
from google.cloud import compute_v1
from google.cloud.monitoring_v3 import Aggregation, MetricServiceClient, TimeInterval, ListTimeSeriesRequest  # noqa: E501

from verinfast.cloud.gcp.zones import zones
from verinfast.cloud.cloud_dataclass import \
    Utilization_Datapoint as Datapoint,  \
    Utilization_Datum as Datum

# os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = "/Users/jason/.config/gcloud/application_default_credentials.json"  # noqa: E501

metric_identifiers = {
    "cpu": 'compute.googleapis.com/instance/cpu/utilization',
    "hdd": 'agent.googleapis.com/disk/percent_used',  # multiple
    "mem": 'agent.googleapis.com/memory/percent_used'
}

metrics = []

now = time.time()
seconds = int(now)
start_seconds = seconds - (30 * 24 * 60 * 60)  # 30 days ago
nanos = int((now - seconds) * 10**9)
interval = TimeInterval(
    {
        "end_time": {"seconds": seconds, "nanos": 0},
        "start_time": {"seconds": start_seconds, "nanos": 0},
    }
)

mean_aggregation = Aggregation(
    {
        "alignment_period": {"seconds": 3600},  # 60 minutes
        "per_series_aligner": Aggregation.Aligner.ALIGN_MEAN,
        "cross_series_reducer": Aggregation.Reducer.REDUCE_MEAN
    }
)

min_aggregation = Aggregation(
    {
        "alignment_period": {"seconds": 3600},  # 60 minutes
        "per_series_aligner": Aggregation.Aligner.ALIGN_MIN,
        "cross_series_reducer": Aggregation.Reducer.REDUCE_MIN
    }
)

max_aggregation = Aggregation(
    {
        "alignment_period": {"seconds": 3600},  # 60 minutes
        "per_series_aligner": Aggregation.Aligner.ALIGN_MAX,
        "cross_series_reducer": Aggregation.Reducer.REDUCE_MAX
    }
)

aggregations = {
    "min": min_aggregation,
    "mean": mean_aggregation,
    "max": max_aggregation
}


def get_metrics_for_instance(sub_id: str, instance_name: str) -> List[Datum]:
    metrics_client = MetricServiceClient()
    results_dict = {}
    for m in metric_identifiers:
        metric = metric_identifiers[m]
        for a in aggregations:
            aggregation = aggregations[a]
            request = ListTimeSeriesRequest(
                filter=f'metric.type = "{metric}" AND metric.labels.instance_name = "{instance_name}"',  # noqa: E501
                view=ListTimeSeriesRequest.TimeSeriesView.FULL,
                aggregation=aggregation,
                name=f"projects/{sub_id}",
                interval=interval
            )
            try:
                results = metrics_client.list_time_series(
                    request=request
                )
                for result in results:
                    for point in result.points:
                        d = str(datetime.fromtimestamp(
                            point.interval.start_time.timestamp()
                        ))
                        if d not in results_dict:
                            results_dict[d] = {}
                        if m not in results_dict[d]:
                            results_dict[d][m] = {}
                        results_dict[d][m][a] = point.value.double_value * 100
                        results_dict[d]["t"] = \
                            point.interval.start_time.timestamp()
            except NotFound:
                pass
    data = []
    for t in results_dict:
        p = results_dict[t]
        temp = {}
        for k in p:
            if k != "t":
                t2 = {}
                if "mean" in p[k]:
                    t2["Average"] = p[k]["mean"]
                else:
                    t2["Average"] = 0
                if "max" in p[k]:
                    t2["Maximum"] = p[k]["max"]
                else:
                    t2["Maximum"] = 0
                if "min" in p[k]:
                    t2["Minimum"] = p[k]["min"]
                else:
                    t2["Minimum"] = 0
                my_datapoint = Datapoint(**t2)
                temp[k] = my_datapoint
        temp["Timestamp"] = p["t"]
        datum = Datum(**temp)
        data.append(datum)
    return data


def get_instances(sub_id: str, path_to_output: str = "./", dry=False):
    if not dry:
        my_instances = []
        # networks_client = compute_v1.NetworksClient()
        instances_client = compute_v1.InstancesClient()
        for zone in zones:
            for instance in instances_client.list(project=sub_id, zone=zone):
                try:
                    name = instance.name
                    architecture = instance.disks[0].architecture
                    hw = instance.machine_type
                    state = instance.status
                    region = zone[0:-3]
                    z = zone[-1:]
                    nic = instance.network_interfaces[0]
                    subnet = nic.subnetwork
                    public_ip = "n/a"
                    if nic.access_configs:
                        public_ip = nic.access_configs[0].nat_i_p
                    vnet_name = nic.network

                    m = get_metrics_for_instance(
                        sub_id=sub_id,
                        instance_name=name
                    )
                    d = {
                            "id": str(instance.id),
                            "metrics": [metric.dict for metric in m]
                        }
                    metrics.append(d)

                    my_instance = {
                        "id": str(instance.id),
                        "name": name,
                        "type": hw.split("/")[-1],
                        "state": state,
                        "zone": z,
                        "region": region,
                        "subnet": subnet.split("/")[-1],
                        "architecture": architecture,
                        "publicIp": public_ip,
                        "vpc": vnet_name.split("/")[-1]
                    }
                except KeyError:
                    continue
                my_instances.append(my_instance)
        # for network in networks_client.list(project=sub_id):
        #     print(network)
        upload = {
                    "metadata": {
                        "provider": "gcp",
                        "account": str(sub_id)
                    },
                    "data": my_instances
                }
    # End dry check
    gcp_output_file = os.path.join(
        path_to_output,
        f'gcp-instances-{sub_id}.json'
    )
    if not dry:
        with open(gcp_output_file, 'w') as outfile:
            outfile.write(json.dumps(upload, indent=4))

        upload['data'] = metrics
        with open(gcp_output_file[:-5]+"-utilization.json", "w") as outfile2:
            outfile2.write(json.dumps(upload, indent=4))

    return gcp_output_file

# Test code
# get_instances(sub_id="startupos-328814")



import json
from unittest.mock import patch
import os
from pathlib import Path
import shutil

from verinfast.agent import Agent
from verinfast.config import Config
from verinfast.upload import Uploader
from verinfast.utils.utils import DebugLog


@patch('verinfast.user.__get_input__', return_value='y')
def test_no_config(self):
    agent = Agent()
    assert agent.config.modules.cloud == []
    assert agent.config.config["local_repos"] == ["./"]


@patch('verinfast.user.__get_input__', return_value='y')
def test_file(self):
    file_path = Path(__file__)
    test_folder = file_path.parent.absolute()
    agent = Agent()
    config = Config()
    config.cfg_path = str(test_folder.joinpath("config.yaml").absolute())
    config.__init__()
    # Test not overwritten
    assert config.cfg_path == str(test_folder.joinpath("config.yaml").absolute())  # noqa: E501
    assert config.modules.cloud == []
    agent.config = config
    assert agent.config.dry is True
    assert agent.config.runDependencies is False
    assert agent.config.reportId == 0
    assert agent.config.baseUrl == ''
    assert agent.up("logs", report=0) == '/report/0/agent_logs'
    agent.uploader = Uploader(agent.config.upload_conf)
    agent.up = agent.uploader.make_upload_path
    assert agent.up("logs", report=0) == '/report/0/agent_logs'
    assert agent.getloc(config.cfg_path) >= 9


@patch('verinfast.user.__get_input__', return_value='y')
def test_str_results_from_file(self):

    file_path = Path(__file__)
    test_folder = file_path.parent.absolute()
    results_dir = test_folder.joinpath("results").absolute()
    agent = Agent()
    config = Config(str(test_folder.joinpath("str_conf.yaml").absolute()))
    config.output_dir = str(results_dir)
    assert config.runGit
    agent.config = config
    agent.debug = DebugLog(path=agent.config.output_dir, debug=False)
    agent.log = agent.debug.log
    assert agent.config.output_dir == str(results_dir)
    assert agent.config.dry is False
    assert agent.config.runGit is True
    try:
        shutil.rmtree(results_dir)
    except Exception as e:
        print(e)
    assert Path(results_dir).exists() is False
    os.makedirs(agent.config.output_dir, exist_ok=True)
    agent.scan()
    assert Path(results_dir).exists() is True
    with open(results_dir.joinpath('small-test-repo.git.filelist.json')) as f:
        file_list = json.load(f)
        assert len(file_list) >= 4
        for myfile in file_list:
            if myfile["name"] == "README.md":
                assert str(myfile["path"]).endswith("README.md")
            assert myfile["name"]
            assert myfile["path"]
    with open(results_dir.joinpath('small-test-repo.git.sizes.json')) as f:
        sizes = json.load(f)
        assert sizes["files"]["."]["size"] >= 34100
        assert sizes["files"]["./README.md"]["ext"] == "md"
    with open(results_dir.joinpath('small-test-repo.git.stats.json')) as f:
        stats = json.load(f)
        for file_name in stats["files"]:
            file_name = str(file_name)
            if file_name.endswith("/temp_repo/README.md"):
                my_file = stats["files"][file_name]
                assert my_file["lang"][0] == "Markdown"
    with open(results_dir.joinpath('small-test-repo.git.findings.json')) as f:
        findings = json.load(f)
        for finding in findings["results"]:
            p = str(finding["path"])
            if p.endswith("semgrep/error.sh"):
                assert finding["start"]["line"] in [5, 10, 16, 19]



from pathlib import Path
import os

import yaml

from typing import List, Callable

user_home = os.path.expanduser('~')


def __get_input__(t: str):
    return input(t)


def repeat_boolean_prompt(
        prompt: str,
        logger: Callable | None = None,
        default_val: bool = False
):
    valid_strs = ["y", "n"]
    print_opts = "(y/N)"
    d = "n"

    if default_val is True:
        print_opts = "(Y/n)"
        d = "y"

    logger(prompt)
    resp = repeat_prompt(prompt=print_opts, valid=valid_strs, default_val=d)

    return resp.lower() == "y"


def repeat_prompt(
        prompt: str,
        valid: List[str],
        default_val: str | None = None
):
    prompt += "\n"
    resp = __get_input__(prompt).lower()
    while resp not in valid:
        if default_val and not resp:
            return default_val
        resp = __get_input__(prompt).lower()
    return resp


def initial_prompt():
    should_upload_log = False

    warning = '''
        VerinFast™ would like to collect information about
        this scan. Do you give consent to upload diagnostic
        information to our sponsor, StartupOS at:
        https://beta.startupos.dev/
    '''

    storage = f'''
        VerinFast™ stores user preferences and configurations
        in your home directory. To do that we create a hidden
        folder (.verinfast) inside the directory {user_home}.

        We never store keys, passwords or confidential information
        there. We do store consents so we don't prompt on every
        run, hashes for more efficient scanning, and other
        preferences.
    '''

    f_path = (
        Path(os.path.expanduser('~/.verinfast/'))
        .joinpath('preferences.yaml')
    )
    conf = {}
    if f_path.exists():
        with open(f_path, "r") as f:
            conf = yaml.safe_load(f)

    if "upload_permission" not in conf:
        resp = repeat_boolean_prompt(warning, print, True)
        if resp:
            conf["upload_permission"] = True
            should_upload_log = True
        else:
            conf["upload_permission"] = False
    else:
        should_upload_log = conf["upload_permission"]

    if "show_warning" not in conf or conf["show_warning"]:
        print(storage)
        conf["show_warning"] = False

    os.makedirs(os.path.expanduser('~/.verinfast'), exist_ok=True)

    with open(f_path, "w") as f:
        f.write(yaml.dump(conf))

    return should_upload_log


def save_path():
    return os.path.expanduser('~/.verinfast/')


if __name__ == "__main__":
    print(initial_prompt())



import datetime
import json
import os

from azure.identity import DefaultAzureCredential
from azure.monitor.query import MetricsQueryClient
from azure.mgmt.storage import StorageManagementClient
from azure.mgmt.resource import ResourceManagementClient


def getBlocks(sub_id: str, path_to_output: str = "./", dry=False):
    if not dry:
        credential = DefaultAzureCredential()
        client = MetricsQueryClient(credential)
        resource_client = ResourceManagementClient(
            credential,
            subscription_id=sub_id
        )

        group_list = resource_client.resource_groups.list()
        storage_client = StorageManagementClient(
            credential,
            subscription_id=sub_id
        )

        known_buckets = {}
        for group in group_list:
            accounts = storage_client.storage_accounts.list_by_resource_group(
                    resource_group_name=group.name
                )

            for account in accounts:
                containers = storage_client.blob_containers.list(
                    resource_group_name=group.name,
                    account_name=account.name
                )

                file_shares = storage_client.file_shares.list(
                    resource_group_name=group.name,
                    account_name=account.name
                )

                # print(containers)
                d = datetime.timedelta(
                    days=1,
                    seconds=0,
                    microseconds=0,
                    milliseconds=0,
                    minutes=0,
                    hours=0,
                    weeks=0
                )

                o = client.query_resource(
                    resource_uri=account.id,
                    metric_names=['UsedCapacity'],
                    timespan=d
                )

                bytes = o.metrics[0].timeseries[0].data[0].average
                print(bytes)
                known_buckets[account.name] = {
                        "name": account.name,
                        "size": bytes,
                        "retention": None,
                        "public": False,
                        "permissions": []
                    }
                for c in containers:
                    if c.public_access:
                        known_buckets[account.name]["public"] = True
                for f in file_shares:
                    print(f)
        my_buckets = list(known_buckets.values())
        upload = {
                    "metadata": {
                        "provider": "azure",
                        "account": str(sub_id)
                    },
                    "data": my_buckets
                }
    # End dry block
    azure_output_file = os.path.join(
        path_to_output,
        f'azure-storage-{sub_id}.json'
    )
    if not dry:
        with open(azure_output_file, 'w') as outfile:
            outfile.write(json.dumps(upload, indent=4))
    return azure_output_file

# Test Code
# getBlocks(sub_id="80dc7a6b-df94-44be-a235-7e7ade335a3c")



import json
import os
from unittest.mock import patch
from pathlib import Path
import shutil

import pytest

from verinfast.agent import Agent
from verinfast.config import Config
from verinfast.utils.utils import DebugLog

file_path = Path(__file__)
test_folder = file_path.parent.absolute()
str_path = str(test_folder.joinpath('str_conf.yaml').absolute())
results_dir = test_folder.joinpath("results").absolute()
saw_message = False
MAX_RECURSION_DEPTH = 10


def check_children(
            i: str | dict | list,
            max_length=30,
            recursion_depth=0,
            # This list must match agent.py
            excludes=[
                "cwe",
                "owasp",
                "path",
                "check_id",
                "license",
                "fingerprint",
                "message",
                "references",
                "url",
                "source",
                "severity"
            ]
        ):
    if recursion_depth > MAX_RECURSION_DEPTH:
        raise Exception("In TOO DEEP!")
    global saw_message
    # print(i)
    if isinstance(i, str):
        if len(i) > max_length:
            print(i)
        assert len(i) <= max_length
    elif isinstance(i, dict):
        for k in i:
            if k == "message":
                saw_message = True
            if k in excludes:
                print(f"{k} is excluded")
            else:
                try:
                    check_children(
                        i[k],
                        recursion_depth=recursion_depth+1,
                        excludes=excludes
                    )
                except Exception as e:
                    print(k)
                    raise e
    elif isinstance(i, list):
        for k in i:
            try:
                check_children(
                    k,
                    recursion_depth=recursion_depth+1,
                    excludes=excludes
                )
            except Exception as e:
                print(k)
                raise e
    elif (
        isinstance(i, float) or
        isinstance(i, int) or
        isinstance(i, bool)
    ):
        pass
    else:
        raise Exception("Non-serializable Object")


@patch('verinfast.user.__get_input__', return_value='y')
def test_no_truncate(self):
    global saw_message
    saw_message = False
    try:
        shutil.rmtree(results_dir)
    except Exception as e:
        print(e)
        pass
    os.makedirs(results_dir, exist_ok=True)
    agent = Agent()
    config = Config(str_path)
    config.output_dir = results_dir
    agent.config = config
    agent.config.dry = False
    agent.config.shouldUpload = False
    agent.debug = DebugLog(path=agent.config.output_dir, debug=False)
    agent.log = agent.debug.log
    agent.scan()
    findings_file_path = results_dir.joinpath('small-test-repo.git.findings.json')  # NOQA: E501

    with open(findings_file_path) as f:
        d = json.load(f)
        r = d["results"]
        assert r[0]["check_id"] == "bash.curl.security.curl-eval.curl-eval"
        m = r[0]["extra"]["message"]
        assert m == "Data is being eval'd from a `curl` command. An attacker with control of the server in the `curl` command could inject malicious code into the `eval`, resulting in a system comrpomise. Avoid eval'ing untrusted data if you can. If you must do this, consider checking the SHA sum of the content returned by the server to verify its integrity."  # noqa: E501
        print("CHECK CHILDREN")
        for k in r:
            with pytest.raises(AssertionError):
                check_children(
                    k,
                    excludes=[
                        "cwe",
                        "path",
                        "check_id",
                        "license",
                        "taint_sink",
                        "taint_source",
                        "fingerprint"
                    ]
                )
    assert saw_message is True


@patch('verinfast.user.__get_input__', return_value='y')
def test_truncate(self):
    global saw_message
    saw_message = False
    try:
        shutil.rmtree(results_dir)
    except Exception as e:
        print(e)
        pass
    os.makedirs(results_dir, exist_ok=True)
    agent = Agent()
    config = Config(str_path)
    config.output_dir = results_dir
    assert config.runGit is True
    assert config.runScan is True
    agent.config = config
    agent.config.dry = False
    agent.config.shouldUpload = False
    agent.config.truncate_findings = True
    agent.config.truncate_findings_length = 0
    assert agent.config.runGit is True
    assert agent.config.runScan is True
    agent.debug = DebugLog(path=agent.config.output_dir, debug=False)
    agent.log = agent.debug.log
    agent.scan()
    findings_file_path = results_dir.joinpath('small-test-repo.git.findings.json')  # NOQA: E501

    with open(findings_file_path) as f:
        d = json.load(f)
        r = d["results"]
        assert r[0]["check_id"] == "bash.curl.security.curl-eval.curl-eval"
        assert r[0]["extra"]["lines"] == ""
        for k in r:
            check_children(k)
    assert saw_message is True


@patch('verinfast.user.__get_input__', return_value='y')
def test_truncate_from_args(self):
    global saw_message
    saw_message = False
    try:
        shutil.rmtree(results_dir)
    except Exception as e:
        print(e)
        pass
    os.makedirs(results_dir, exist_ok=True)
    agent = Agent()
    config = Config(str_path)
    parser = config.init_argparse()
    args = parser.parse_args(['--truncate_findings=30'])
    assert args.truncate_findings == 30
    config.handle_args(args)
    assert config.runGit is True
    assert config.runScan is True
    assert config.truncate_findings is True
    assert config.truncate_findings_length == 30
    config.output_dir = results_dir
    agent.config = config
    assert agent.config.truncate_findings is True
    assert agent.config.truncate_findings_length == 30
    agent.config.dry = False
    agent.config.shouldUpload = False
    agent.debug = DebugLog(path=agent.config.output_dir, debug=False)
    agent.log = agent.debug.log
    agent.scan()
    findings_file_path = results_dir.joinpath('small-test-repo.git.findings.json')  # NOQA: E501

    with open(findings_file_path) as f:
        d = json.load(f)
        r = d["results"]
        assert r[0]["check_id"] == "bash.curl.security.curl-eval.curl-eval"
        for k in r:
            check_children(k)
    assert saw_message is True



import json
import xml.etree.ElementTree as ET

from verinfast.dependencies.walkers.classes import Walker, Entry


class NuGetWalker(Walker):
    def initialize(self, command: str = None):
        discoveryUrl = 'https://api.nuget.org/v3/index.json'
        uselessList = json.loads(
            self.getUrl(discoveryUrl)
        )
        for r in uselessList["resources"]:
            if "Catalog" in r["@type"]:
                self.catalogUrl = r["@id"]
            elif r["@type"] == "RegistrationsBaseUrl":
                self.registrationUrl = r["@id"]
        if command:
            return super().initialize(command)

    def get_license(self, name: str, version: str) -> str:
        resp = None
        name2 = name.lower()
        try:
            license_resp = self.getUrl(f"{self.registrationUrl}{name}/{version}.json")  # NOQA:E501
            resp = json.loads(license_resp)
        except Exception as e:
            self.log(e, display=False)
            try:
                license_resp = (
                    self
                    .getUrl(f"{self.registrationUrl}{name2}/{version}.json")
                )
                resp = json.loads(license_resp)
            except Exception as e2:
                self.log(
                    f"License not found for: {name}@{version}",
                    display=True
                )
                self.log(e2, display=False)
        if resp is not None:
            catalog_entry_url = resp["catalogEntry"]
            catalog_entry = json.loads(self.getUrl(catalog_entry_url))
            if "licenseExpression" in catalog_entry:
                return catalog_entry["licenseExpression"]
            elif "licenseUrl" in catalog_entry:
                return catalog_entry["licenseUrl"]
            else:
                return ""
        else:
            return ""

    def parse(self, file: str, expand=False):
        # Reference: <PackageReference Include="Swashbuckle.AspNetCore" Version="6.5.0" /> # noqa: E501
        tree = ET.parse(file)
        x_path = "ItemGroup/PackageReference"
        dependencies = tree.findall(x_path)
        for d in dependencies:
            v = d.attrib["Version"]
            n = d.attrib["Include"]
            lic = self.get_license(name=n, version=v)
            e = Entry(
                source="nuget",
                name=n,
                specifier="=="+v,
                license=lic,
            )
            self.entries.append(e)


# Example: https://github.com/umbraco/Umbraco-CMS/pull/13787/files
c_sharp_matches = ["*.csproj"]



import http.client
import json
import os
import re
import ssl
import subprocess

from verinfast.utils.utils import DebugLog
debugLog = DebugLog(os.getcwd())


def runAzure(subscription_id, start, end, path_to_output, dry=False):
    if not dry:
        ssl.SSLContext.verify_mode = property(
            lambda self: ssl.CERT_OPTIONAL,
            lambda self,
            newval: None
        )

        conn = http.client.HTTPSConnection("management.azure.com")
        subprocess.run(
            f'az account set --subscription {subscription_id}',
            shell=True
        )
        # results = subprocess.run(f'az account show --query id" -o tsv', shell=True, stdout=subprocess.PIPE)  # noqa: E501
        # subscription_id = results.stdout.decode()[:-1]
        # 80dc7a6b-df94-44be-a235-7e7ade335a3c
        req_path = f'/subscriptions/{subscription_id}/providers/Microsoft.CostManagement/query'  # noqa: E501
        req_params = "api-version=2023-03-01"
        body = {
            "dataset": {
                "granularity": "Daily",
                "aggregation": {
                    "totalCost": {
                        "name": "PreTaxCost",
                        "function": "Sum"
                    }
                },
                "grouping": [
                    {
                        "type": "Dimension",
                        "name": "ServiceName"
                    }
                ]
            },
            "timeframe": "custom",
            "timePeriod": {
                    "from": str(start),
                    "to": str(end)
            },
            "type": "ActualCost"
        }
        results = subprocess.run(
            "az account get-access-token --resource=https://management.azure.com/ --query accessToken -o tsv",  # noqa: E501
            shell=True,
            stdout=subprocess.PIPE
        )

        bearer_token = "Bearer "+results.stdout.decode().strip()
        # print(bearer_token)
        print("Fetching data for subscription: " + subscription_id)
        conn.request(
            "POST",
            req_path + "?" + req_params,
            headers={
                "Authorization": bearer_token,
                "Content-Type": "application/json"
            },
            body=json.dumps(body)
        )
        response = conn.getresponse()
        print("Parsing response")
        data = json.loads(response.read().decode())
        if 'error' in data:
            debugLog.log(msg=json.dumps(data["error"]), tag="Azure error")
            return

        with open(f"{path_to_output}/azure_output_raw.json", "w") as outfile:
            outfile.write(json.dumps(data, indent=4))

        vals = data["properties"]["rows"]
        new_vals = []
        for val in vals:
            nv = {
                "Date": str(val[1]),
                "Cost": str(val[0]),
                "Group": val[2],
                "Currency": val[3]
            }
            s = nv["Date"]
            if re.match('^[0-9]*$', s) and len(s) == 8:
                nv["Date"] = s[0:4] + "-" + s[4:6] + "-" + s[6:8]
            new_vals.append(nv)
        upload = {
            "metadata": {
                "provider": "azure",
                "account": subscription_id
            },
            "data": new_vals
        }
    # End dry block

    az_output_file = os.path.join(
        path_to_output,
        f'az-cost-{subscription_id}.json'
    )

    if not dry:
        with open(az_output_file, "w") as outfile:
            outfile.write(json.dumps(upload, indent=4))

    return az_output_file



import json
import os
import subprocess

from verinfast.utils.utils import DebugLog


debugLog = DebugLog(os.getcwd())


def find_profile(targeted_account: str, log=debugLog.log):
    profiles = []
    available_accounts = []
    results = subprocess.run(
        "aws configure list-profiles",
        shell=True,
        stdout=subprocess.PIPE
    )
    text = results.stdout.decode()
    for line in text.splitlines():
        profiles.append(line)
        cmd = f'aws sts get-caller-identity --profile="{line}" --output=json'
        try:
            results = subprocess.run(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
                check=True
            )
        except subprocess.CalledProcessError:
            # TODO make some log message that makes sense
            continue
        text = results.stdout.decode()
        identity = json.loads(text)
        account = identity["Account"]
        available_accounts.append(account)
        if str(account) == str(targeted_account):
            return line

    debugLog.log(msg=profiles, tag="AWS Profiles")
    debugLog.log(msg=available_accounts, tag="AWS Available Accounts")
    return None



from datetime import datetime, timedelta
import json
import os
from typing import List

import boto3
import botocore

import verinfast.cloud.aws.regions as r
from verinfast.cloud.cloud_dataclass import (
    Utilization_Datapoint as Datapoint,
    Utilization_Datum as Datum
)

regions = r.regions


def get_metric_for_instance(
            metric: str,
            instance_id: str,
            session,
            region: str,
            namespace: str = 'AWS/EC2',
            unit: str = 'Percent',
        ):
    try:
        client = session.client('cloudwatch', region_name=region)
        response = client.get_metric_statistics(
            Namespace=namespace,
            MetricName=metric,
            Dimensions=[
                {
                    'Name': 'InstanceId',
                    'Value': instance_id
                },
            ],
            StartTime=datetime.utcnow() - timedelta(days=30),
            EndTime=datetime.utcnow(),
            Period=60*60,
            Statistics=[
                'Average',
                'Maximum',
                'Minimum'
            ],
            Unit=unit
        )
        return response
    except botocore.exceptions.ClientError:
        pass


def parse_multi(datapoint: dict | List[dict]) -> Datapoint:
    dp_sum: float = 0
    dp_count: int = 0
    dp_min: float = 0
    dp_max: float = 0

    my_datapoint = Datapoint()
    my_datapoint.Timestamp = datapoint['Timestamp']
    if type(datapoint) is not list:
        datapoint = [datapoint]
    for entry in datapoint:
        if 'Average' in entry:
            e = entry['Average']
            dp_sum += e
            dp_count += 1
        if 'Minimum' in entry:
            dp_min += entry['Minimum']
        if 'Maximum' in entry:
            dp_max += entry['Maximum']
    dp_count = max(dp_count, 1)
    my_datapoint.Average = dp_sum / dp_count
    my_datapoint.Maximum = dp_max / dp_count
    my_datapoint.Minimum = dp_min / dp_count

    return my_datapoint


def get_instance_utilization(
            instance_id: str,
            session,
            region: str
        ) -> List[Datum]:
    cpu_resp = get_metric_for_instance(
            metric='CPUUtilization',
            instance_id=instance_id,
            session=session,
            region=region
        )
    mem_resp = get_metric_for_instance(
            metric='mem_used_percent',
            instance_id=instance_id,
            namespace='CWAgent',
            session=session,
            region=region
        )

    hdd_resp = get_metric_for_instance(
            metric='disk_used_percent',
            instance_id=instance_id,
            namespace='CWAgent',
            session=session,
            region=region
        )

    cpu_stats: List[Datapoint] = []
    mem_stats: List[Datapoint] = []
    hdd_stats: List[Datapoint] = []

    # each instance may have more than 1 CPU
    for datapoint in cpu_resp['Datapoints']:
        summary = parse_multi(datapoint)
        cpu_stats.append(summary)

    # memory and disk are not collected by default
    # each instance may have more than one disk
    if "Datapoints" in hdd_resp:
        for datapoint in hdd_resp["Datapoints"]:
            summary = parse_multi(datapoint)
            hdd_stats.append(summary)
    # memory
    if "Datapoints" in mem_resp:
        mem_stats = [Datapoint.From(i) for i in mem_resp["Datapoints"]]

    data = []
    m = max(len(cpu_stats), len(mem_stats), len(hdd_stats))
    if m == 0:
        return []
    else:
        for i in range(m):
            datum = Datum(Timestamp=cpu_stats[i].Timestamp)
            if i < len(cpu_stats):
                datum.cpu = cpu_stats[i]
            else:
                datum.cpu = None
            if i < len(mem_stats):
                datum.mem = mem_stats[i]
            else:
                datum.mem = None
            if i < len(hdd_stats):
                datum.hdd = hdd_stats[i]
            else:
                datum.hdd = None
            data.append(datum)
    return data


def get_instances(sub_id: int, path_to_output: str = "./",
                  dry=False) -> str | None:
    if not dry:
        session = boto3.Session()
        profiles = session.available_profiles
        right_session = None
        for profile in profiles:
            s2 = boto3.Session(profile_name=profile)
            sts = s2.client('sts')
            try:
                id = sts.get_caller_identity()
            except botocore.exceptions.ClientError:
                continue
            if str(id['Account']) == str(sub_id):
                right_session = s2
                break
        if right_session is None:
            return None
        my_instances = []
        metrics = []
        for region in regions:
            try:
                client = right_session.client('ec2', region_name=region)
                paginator = client.get_paginator('describe_instances')
                page_iterator = paginator.paginate()
                for page in page_iterator:
                    reservations = page['Reservations']
                    for reservation in reservations:
                        instances = reservation['Instances']
                        for instance in instances:
                            tags = instance['Tags']
                            if tags is None:
                                name = instance["InstanceId"]
                            else:
                                tags_with_name = [t['Value'] for t in tags if t['Key'] == 'Name']  # noqa: E501
                                if not tags_with_name:
                                    name = instance["InstanceId"]
                                else:
                                    name = tags_with_name[0]
                            m = get_instance_utilization(
                                instance_id=instance["InstanceId"],
                                session=right_session,
                                region=region
                            )
                            d = {
                                    "id": instance["InstanceId"],
                                    "metrics": [metric.dict for metric in m]
                                }
                            metrics.append(d)
                            region = None
                            subnet_id = None
                            zone = None
                            if 'SubnetId' in instance:
                                subnet_id = instance['SubnetId']
                            if 'Placement' in instance:
                                placement = instance['Placement']
                                if 'AvailabilityZone' in placement:
                                    zone = placement['AvailabilityZone']
                                    region = zone[0:-1]
                            try:
                                result = {
                                    "id": instance["InstanceId"],
                                    "name": name,
                                    "state": instance["State"]["Name"],
                                    "type": instance['InstanceType'],
                                    "zone": zone,
                                    "region": region,
                                    "subnet": subnet_id,
                                    "architecture": instance['Architecture']
                                }
                            except KeyError:
                                continue
                            if "VpcId" in instance:
                                result["vpc"] = instance['VpcId']
                            else:
                                result["vpc"] = 'n/a'
                            if "PublicIpAddress" in instance:
                                result["publicIp"] = instance['PublicIpAddress']  # noqa: E501
                            else:
                                result["publicIp"] = 'n/a'
                            if "NetworkInterfaces" in instance:
                                ni = instance["NetworkInterfaces"]
                                for interface in ni:
                                    if 'Association' in interface:
                                        association = interface['Association']
                                        if 'PublicIp' in association:
                                            public_ip = association['PublicIp']
                                            result["publicIp"] = public_ip
                            my_instances.append(result)
            except botocore.exceptions.ClientError:
                pass

        upload = {
                    "metadata": {
                        "provider": "aws",
                        "account": str(sub_id)
                    },
                    "data": my_instances
                }
    # End dry check

    aws_output_file = os.path.join(
        path_to_output,
        f'aws-instances-{sub_id}.json'
    )

    if not dry:
        with open(aws_output_file, 'w') as outfile:
            outfile.write(json.dumps(upload, indent=4))
        upload['data'] = metrics
        with open(aws_output_file[:-5]+"-utilization.json", "w") as outfile2:
            outfile2.write(json.dumps(upload, indent=4))

    return aws_output_file



import json
import os
import subprocess
from pathlib import Path
from typing import List

from curses.ascii import isdigit

from verinfast.dependencies.walkers.classes import Walker, Entry


class NodeWalker(Walker):
    def initialize(self, root_path: str = "./"):
        root_path = str(Path(root_path).absolute())
        self.install_points: List[Path] = []
        for p in Path(root_path).rglob('**/*.*'):
            if p.name in self.manifest_files:
                self.install_points.append(p)
        for p in self.install_points:
            target_dir = Path(p).parent
            os.chdir(target_dir)
            try:
                subprocess.run(
                    "npm install --production --silent --yes",
                    capture_output=True,
                    shell=True
                )
            except Exception as error:
                self.log(f'Error with npm install: {error}')
            else:
                self.walk('node_modules')
            os.chdir(root_path)

    def parse(self, file: str, expand=False):
        entry = {}
        license = {}
        try:
            with open(file) as f:
                d = json.load(f)
                if "name" in d and "version" in d:
                    key = d["name"] + "@" + d["version"]
                    entry["source"] = "npm"
                    entry["name"] = d["name"]
                    entry["specifier"] = "==" + d["version"]
                    if "license" in d and isinstance(d["license"], dict):
                        license[key] = d["license"]["type"]
                    elif "license" in d and isinstance(d["license"], list):
                        license[key] = ' '.join(d["license"])
                    elif "license" in d:
                        license[key] = d["license"]
                    else:
                        license[key] = ""
                    entry["license"] = license[key]
                    entry["requires"] = []
                    if "dependencies" in d:
                        for key in d["dependencies"].keys():
                            k = key
                            value = d["dependencies"][k]
                            if isdigit(value[0]):
                                value = "==" + value
                            entry["requires"].append(k+value)

                    entry["required_by"] = []
                    if "description" in d:
                        entry["summary"] = d["description"]
                    else:
                        entry["summary"] = ""

                    e = Entry(
                        name=entry["name"],
                        specifier=entry["specifier"],
                        source=entry["source"],
                        license=entry["license"],
                        summary=entry["summary"]
                    )
                    self.entries.append(e)
        except Exception as error:
            # handle the exception
            self.log(f"error parsing {file}")
            self.log(error)



from dataclasses import dataclass
from datetime import datetime
from typing import Optional


@dataclass
class Utilization_Datapoint():
    Minimum: Optional[float] = None
    Average: Optional[float] = None
    Maximum: Optional[float] = None
    Timestamp: Optional[datetime] = None

    @property
    def dict(self):
        o = {}
        if (
            self.Minimum is None and
            self.Average is None and
            self.Maximum is None
        ):
            return None

        if self.Minimum is not None:
            o["minimum"] = self.Minimum
        if self.Average is not None:
            o["average"] = self.Average
        if self.Maximum is not None:
            o["maximum"] = self.Maximum
        return o

    @staticmethod
    def From(i: dict):
        if {"Minimum", "Average", "Maximum"} <= i.keys():
            return Utilization_Datapoint(
                Minimum=i["Minimum"],
                Average=i["Average"],
                Maximum=i["Maximum"])
        else:
            raise Exception("Invalid Datapoint")


@dataclass
class Utilization_Datum():
    Timestamp: datetime | float
    cpu: Optional[Utilization_Datapoint] = None
    mem: Optional[Utilization_Datapoint] = None
    hdd: Optional[Utilization_Datapoint] = None

    @property
    def dict(self):
        if type(self.Timestamp) is datetime:
            s = int(self.Timestamp.timestamp())
        else:
            s = self.Timestamp
        o = {"timestamp": s}
        if self.cpu is not None:
            o["cpu"] = self.cpu.dict
        if self.mem is not None:
            o["mem"] = self.mem.dict
        if self.hdd is not None:
            o["hdd"] = self.hdd.dict
        return o



from datetime import datetime, timedelta

import json
import os

import boto3
import botocore.exceptions

import verinfast.cloud.aws.regions as r

regions = r.regions


def getBlocks(sub_id: str, log, path_to_output: str = "./", dry=False):

    if not dry:
        session = boto3.Session()
        profiles = session.available_profiles
        right_session = None
        for profile in profiles:
            s2 = boto3.Session(profile_name=profile)
            sts = s2.client('sts')
            try:
                id = sts.get_caller_identity()
            except botocore.exceptions.ClientError:
                continue
            if int(id['Account']) == int(sub_id):
                right_session = s2
                break
        if right_session is None:
            return None
        # for region in regions:
        #     try:
        s3 = right_session.client('s3', region_name=regions[0])
        response = s3.list_buckets()
        known_buckets = {}
        for bucket in response['Buckets']:
            bucket_name = bucket["Name"]
            resp = s3.get_bucket_location(Bucket=bucket_name)

            permissions = []
            public = False
            try:
                policy_status_resp = s3.get_bucket_policy_status(Bucket=bucket_name)
                public = policy_status_resp["PolicyStatus"]["IsPublic"]

                if "Policy" in policy_status_resp:
                    p_string = policy_status_resp["Policy"]
                    p_dict = json.loads(p_string)
                    statements = p_dict["Statement"]
                    # print(statements)
                    permissions = [json.dumps(s) for s in statements]
                    # print(s2)
            except s3.exceptions.from_code('NoSuchBucketPolicy'):
                log(msg=bucket_name, tag="No Bucket Policy for bucket")

            versioning = None
            try:
                versioning_response = s3.get_bucket_versioning(Bucket=bucket_name)
                if "Status" in versioning_response:
                    versioning = versioning_response["Status"]

            except s3.exceptions.from_code('NoSuchBucketStatus'):
                log(msg=bucket_name, tag="No Bucket Status for bucket")

            region = resp['LocationConstraint']
            # print(region)
            if region:
                # print('Have region')
                cloudwatch = right_session.client('cloudwatch', region_name=region)
            else:
                cloudwatch = right_session.client(
                    'cloudwatch',
                    region_name='us-east-1'
                )

            response = cloudwatch.get_metric_statistics(
                Namespace="AWS/S3",
                MetricName="BucketSizeBytes",
                Dimensions=[
                    {
                        u"Name": u"BucketName",
                        u"Value": bucket_name
                    },
                    {
                        u'Name': 'StorageType',
                        u'Value': 'StandardStorage'
                    }
                ],
                StartTime=datetime.now() - timedelta(days=2),
                EndTime=datetime.now(),
                Period=3600,
                Statistics=['Average'],
                Unit="Bytes"
            )
            if response['Datapoints']:
                bucket_size_bytes = response['Datapoints'][-1]['Average']
                known_buckets[bucket_name] = {
                    "name": bucket_name,
                    "size": int(bucket_size_bytes),
                    "retention": versioning,
                    "public": public,
                    "permissions": permissions
                }
            else:
                pass
                # print(response)
            # except:
            #     pass

        my_buckets = list(known_buckets.values())
        upload = {
                    "metadata": {
                        "provider": "aws",
                        "account": str(sub_id)
                    },
                    "data": my_buckets
                }
    # End dry block

    aws_output_file = os.path.join(
        path_to_output,
        f'aws-storage-{sub_id}.json'
    )

    if not dry:
        with open(aws_output_file, 'w') as outfile:
            outfile.write(json.dumps(upload, indent=4))

    return aws_output_file

# Test Code
# getBlocks(sub_id='436708548746')



import json
import os
from pathlib import Path
from typing import List

from verinfast.dependencies.walkers.classes import Walker, Entry
from verinfast.utils.utils import std_exec


class ComposerWalker(Walker):
    dependency_tree = {}

    def initialize(self, root_path: str = "./"):
        root_path = str(Path(root_path).absolute())
        self.install_points: List[Path] = []
        for p in Path(root_path).rglob('**/*.*'):
            if p.name in self.manifest_files:
                self.install_points.append(p)
        for p in self.install_points:
            target_dir = Path(p).parent
            os.chdir(target_dir)
            try:
                res = std_exec([
                    "composer",
                    "install",
                    "--no-dev",
                    "--no-progress"
                ])
                self.log(tag="", msg=res, timestamp=False)
            except Exception as error:
                self.log(
                    tag="ERROR",
                    msg=f'Error with composer install: {error}'
                )
            finally:
                self.walk('./')
            for dep in self.dependency_tree:
                dependency = self.dependency_tree[dep]
                for v in dependency:
                    version = dependency[v]
                    self.entries.append(version)
            os.chdir(root_path)

    def parse(self, file: str, expand=False):
        try:
            with open(file) as f:
                cjson = json.load(f)
                if (
                    "license" in cjson and
                    self.dependency_tree.get(cjson['name'])
                ):
                    for version in self.dependency_tree[cjson['name']]:
                        self.dependency_tree[cjson['name']
                                             ][version].license = cjson[
                                                 'license']
                        if "description" in cjson:
                            self.dependency_tree[cjson['name']
                                                 ][version].summary = cjson[
                                                     'description']
                if 'require' in cjson:
                    for k, v in cjson['require'].items():
                        if self.dependency_tree.get(k):
                            self.dependency_tree[k] = self.dependency_tree[k]
                        else:
                            self.dependency_tree[k] = {}
                        if self.dependency_tree[k].get(v):
                            self.dependency_tree[k][v] = \
                                self.dependency_tree[k][v]
                        else:
                            self.dependency_tree[k][v] = Entry(
                                name=k, source="composer", specifier=v)
                self.log(self.dependency_tree)
        except Exception as error:
            self.log(f"error parsing {file}")
            try:
                self.log(json.load(file)['name'])
                self.log(json.load(file)['require'])
            except Exception as ex:
                raise ex
            self.log(error)
